/******************************************************************
* Reconstruccion algebraica de volumenes usando el metodo SART
* ------------------------------------------------------------
* Programa para reconstruir volúmenes a partir de una colección de
* proyecciones suyas. 
*
* Proyecto de Fin de Carrera, Universidad Complutense de Madrid
* Facultad de Informatica, 2006/2007
*
* David Garcia Soriano
* Enrique Martin Martin
* David Romero Laorden
******************************************************************/

#ifndef __SART_H
#define __SART_H

typedef double nreal;
#define eps 1e-7
#define maxtam 200         // m�imo tama� de una imagen

typedef nreal tipo;
struct punto {
  tipo x, y, z;
  punto(matrix1D<double> m) : x(XX(m)), y(YY(m)), z(ZZ(m)) {} // implicit casting 
  punto(tipo a = 0, tipo b = 0, tipo c = 0) : x(a), y(b), z(c) {}
  punto operator +(punto b) const { return punto(x + b.x, y + b.y, z + b.z); }
  punto operator -(punto b) const { return punto(x - b.x, y - b.y, z - b.z); }
  punto operator *(tipo r) const { return punto(x * r, y * r, z * r); }
  punto operator /(tipo r) const { return punto(x / r, y / r, z / r); }
  friend punto operator *(tipo r, punto p) { return punto(p.x * r, p.y * r, p.z * r); }
  friend punto operator /(tipo r, punto p) { return punto(p.x / r, p.y / r, p.z / r); }
  void operator +=(punto b) { x += b.x; y += b.y; z += b.z; }
  void operator -=(punto b) { x -= b.x; y -= b.y; z -= b.z; }
  bool operator ==(punto p) const { return fabs(x - p.x) < eps && fabs(y - p.y) < eps && fabs(z - p.z) < eps; }
  bool operator !=(punto p) const { return !(fabs(x - p.x) < eps && fabs(y - p.y) < eps && fabs(z - p.z) < eps); }
  friend ostream &operator<<(ostream &out, punto p) { return out << p.x << " " << p.y << " " << p.z << endl; }

  // Producto escalar
  tipo operator *(punto b) const { return x * b.x + y * b.y + z * b.z; }

  tipo mod2() const { return x * x + y * y + z * z; }
  nreal mod() const { return sqrt(x * x + y * y + z * z); }
  tipo distcuad(punto b) const {
    tipo dx = b.x - x, dy = b.y - y, dz = b.z - z;
    return dx * dx + dy * dy + dz * dz;
  }

  /* Distancia a otro punto */
  nreal dist(punto b) const { return sqrt(distcuad(b)); }
  
  void normaliza() {
    tipo d = (tipo)mod();
    x /= d; y /= d; z /= d;
  }
};

struct coordvoxel {
  int x, y, z;
  nreal peso;
  coordvoxel(int xx = 0, int yy = 0, int zz = 0, nreal p = 0) : x(xx), y(yy), z(zz), peso(p) {}
};


#endif 
