
#include <XmippData/xmippImages.hh>
#include <XmippData/xmippSelFiles.hh>
#include <XmippData/xmippMatrices3D.hh>
#include <XmippData/xmippMatrices2D.hh>
#include <XmippData/xmippMatrices1D.hh>
#include <XmippData/xmippGeometry.hh>
#include <XmippData/xmippVolumes.hh>
#include <XmippData/xmippProjection.hh>
#include <iostream>
#include <cassert>
#include <cstring>
#include <cmath>
#include <algorithm>
#include "comun.h"
using namespace std;


int main( int argc, char** argv){
  FileName fnVol1, fnVol2;
  VolumeXmipp vol1, vol2;

  if (argc < 3) { puts("Usage:\ndifvol [theoretical volume] [reconstructed volume]"); return -1; }

  fnVol1 = argv[1];
  fnVol2 = argv[2];
  int Zdim, Ydim, Xdim,
      Zdim2, Ydim2, Xdim2;
  try {
    vol1.read(fnVol1);
    vol2.read(fnVol2);

    Zdim = vol1.img.SliNo(); Ydim = vol1.img.RowNo(); Xdim = vol1.img.ColNo();
    Zdim2 = vol2.img.SliNo(); Ydim2 = vol2.img.RowNo(); Xdim2 = vol2.img.ColNo();

    if (Zdim != Zdim2 || Ydim != Ydim2 || Xdim != Xdim2) {
      puts("Volume sizes do not match");
      printf("theoretical (ZxYxX): %i %i %i\nreconstructed (ZxYxX): %i %i %i\n", Zdim, Ydim, Xdim,
          Zdim2, Ydim2, Xdim2);
      return -2;
    }
  }catch (Xmipp_error XE) {cout << XE; exit(1);}

  int z, y, x, tamano = Zdim * Ydim * Xdim, vacio[2] = { 0 };
  double *vol[2];
  vol[0] = (double *)malloc(tamano * sizeof(*vol[0]));
  vol[1] = (double *)malloc(tamano * sizeof(*vol[1]));
  int pos = 0, i;
  for (z = 0; z < Zdim; ++z) for (y = 0; y < Ydim; ++y) for (x = 0; x < Xdim; ++x, ++pos) {
    vol[0][pos] = vol1(z, y, x);
    vol[1][pos] = vol2(z, y, x);
  }

/*
  double aa[] = { 1,  -0.3, 4.5, 8, 9 },
         bb[] = { -3, -5.5, 4,  11, 13 };
  double *vol[2];
  int tamano = 5, i, pos;
  vol[0] = (double *)malloc(tamano * sizeof(*vol[0]));
  vol[1] = (double *)malloc(tamano * sizeof(*vol[1]));
  for (i = 0; i < tamano; ++i) {    
    vol[0][i] = aa[i];
    vol[1][i] = bb[i];
  }
*/
  // Normaliza el volumen terico a [0..1], simplemente para intentar dar un valor "absoluto" al resultado final
  int n = tamano;
  double minimo = *min_element(&vol[0][0], &vol[0][n]),
         maximo = *max_element(&vol[0][0], &vol[0][n]),
         rango = maximo - minimo;
  for (i = 0; i < n; ++i)
    vol[0][i] = (vol[0][i] - minimo) / rango;

  // Calcula la mejor transformacin lineal que pasa de vol2 a vol1
  double sxi = 0, syi = 0, sxiyi = 0, sxi2 = 0;
  for (i = 0; i < tamano; ++i) {
    sxi += vol[1][i];
    syi += vol[0][i];
    sxiyi += vol[0][i] * vol[1][i];
    sxi2 += vol[1][i] * vol[1][i];
  }
  double a = (n * sxiyi - sxi * syi) / (n * sxi2 - sxi * sxi),
         b = (syi - a * sxi) / n;

  // Transforma el segundo volumen segn esa recta
  for (i = 0; i < tamano; ++i)
    vol[1][i] = a * vol[1][i] + b;

  double error = 0;
  for (pos = 0; pos < tamano; ++pos) {
    double dif = vol[0][pos] - vol[1][pos];
    error += dif * dif;
  }
  error /= tamano;
  printf("a = %lf\tb = %lf\n", a, b);
  printf("Mean squared error after best linear fit: %lf\n", error);

  free(vol[0]);
  free(vol[1]);
  return 0;
}

