/******************************************************************
* Reconstruccion algebraica de volumenes usando el metodo SART
* ------------------------------------------------------------
* Programa para reconstruir volúmenes a partir de una colección de
* proyecciones suyas. 
*
* Proyecto de Fin de Carrera, Universidad Complutense de Madrid
* Facultad de Informatica, 2006/2007
*
* David Garcia Soriano
* Enrique Martin Martin
* David Romero Laorden
******************************************************************/

#ifndef __PUNTO_H
#define __PUNTO_H


typedef nreal tipo;
struct punto;
__device__ punto ppunto(punto p);
__device__ punto ppunto(float, float, float);

struct punto {
  tipo x, y, z;

  __device__ void assign(tipo a = 0, tipo b = 0, tipo c = 0) { x = a; y = b; z = c; }
  __device__ void assign(punto p) { x = p.x; y = p.y; z = p.z; }
  __device__ punto operator +(punto b) const { return ppunto(x + b.x, y + b.y, z + b.z); }
  __device__ punto operator -(punto b) const { return ppunto(x - b.x, y - b.y, z - b.z); }
  __device__ punto operator *(tipo r) const { return ppunto(x * r, y * r, z * r); }
  __device__ punto operator /(tipo r) const { return ppunto(x / r, y / r, z / r); }
  __device__ void operator +=(punto b) { x += b.x; y += b.y; z += b.z; }
  __device__ void operator -=(punto b) { x -= b.x; y -= b.y; z -= b.z; }
  __device__ bool operator ==(punto p) const { return fabs(x - p.x) < eps && fabs(y - p.y) < eps && fabs(z - p.z) < eps; }
  __device__ bool operator !=(punto p) const { return !(fabs(x - p.x) < eps && fabs(y - p.y) < eps && fabs(z - p.z) < eps); }
 
  // Producto escalar
  __device__ tipo operator *(punto b) const { return x * b.x + y * b.y + z * b.z; }

  // Mdulo del producto vectorial
  __device__ tipo operator ^(punto b) const {
    float xx = y * b.z - z * b.y, yy = b.x * z - x * b.z, zz = x * b.y - b.x * y;
    return sqrt(xx * xx + yy * yy + zz * zz);
  }

  __device__ tipo mod2() const { return x * x + y * y + z * z; }
  __device__ nreal mod() const { return sqrt(x * x + y * y + z * z); }
  __device__ tipo distcuad(punto b) const {
    tipo dx = b.x - x, dy = b.y - y, dz = b.z - z;
    return dx * dx + dy * dy + dz * dz;
  }

  /* Distancia a otro punto */
 __device__  nreal dist(punto b) const { return sqrt(distcuad(b)); }
};

  // simulacros de constructor
__device__ punto ppunto(punto p) { punto q; q.assign(p); return q; }
__device__ punto ppunto(float x, float y, float z) { punto q; q.assign(x, y, z); return q; }


#endif 
