/* vim: set filetype=cpp : */
#include "stolen.cu"

/* 85% del tiempo total
 * Optimizaciones:
 * - reordenar (pos)? para que accedan a posiciones consecutivas de memoria (hecho)
 * - pasar solo dos filas de la eulerMatrix (hecho)
 * - hacer una funcion para cada valor de dir (no vale la pena)
 *
 *   k es posbloque */
__global__ void proyectavoxels(const float *volumen, tiporayosporvoxel *rayosporvoxel, infoproyslice *slice,
const int Zdim, const int Ydim, const int Xdim,
float a, float b, float c, float d, float e, float f,
int dir, int k, int hbproyecta   
#ifdef DEBUG
, float *debug
#endif
) {
  const int midX = (Xdim + 1) / 2, midY = (Ydim + 1) / 2, midZ = (Zdim + 1) / 2;  
  int pos = blockIdx.x * hbproyecta + threadIdx.x, numslice, x, y, z; 

  /* 0: midZ * midY * Xdim  (eje x)
   * 1: midZ * Ydim * midX  (eje y)
   * 2: Zdim * midY * midX  (eje z) */
  switch (dir) {  // switch falso, todos los hilos eligen el mismo, no afecta al tiempo de ejecucion
    case 0: z = ((pos / (midY * Xdim)) << 1) | (k >> 1); y = (((pos / Xdim) % midY) << 1) | (k & 1); x = pos % Xdim; 
            numslice = x; break;
    case 1: z = ((pos / (Ydim * midX)) << 1) | (k >> 1); y = (pos / midX) % Ydim; x = ((pos % midX) << 1) | (k & 1); 
            numslice = y; break;
    case 2: z = pos / (midY * midX); y = (((pos / midX) % midY) << 1) | (k >> 1); x = ((pos % midX) << 1) | (k & 1);
            numslice = z;
  }
  if (z >= Zdim || y >= Ydim || x >= Xdim) return;
  pos = pos3d(z, y, x);
  slice += Ydim * Xdim * numslice;
  
  // Calcula la proyeccion del centro del voxel
  float projx = a * (x - midX) + b * (y - midY) + c * (z - midZ) + midX,
        projy = d * (x - midX) + e * (y - midY) + f * (z - midZ) + midY;

  // El centro del voxel cae en (projy, projx), que esta en el pixel (pixely, pixelx)
  int pixely = roundf(projy), pixelx = roundf(projx), 
      dira = pixely <= projy ? -1 : 1, dirb = pixelx <= projx ? -1 : 1;
  float dify = projy - pixely, difx = projx - pixelx;

  // Manejamos el vector a mano (speedup 25%)
  int tam = 0; inforayo *array = rayosporvoxel[pos].array;
  float volpos = volumen[pos];

  // hasta aqui el 28% de la proyeccion
  
  // Mira los 4 pixeles (rayos) candidatos a cortar a la esfera
  for (int a = 0, na = 0; na < 2; a += dira, ++na)
    for (int b = 0, nb = 0; nb < 2; b += dirb, ++nb) {
      float da = dify + a, db = difx + b, dist2 = da * da + db * db;

      /* Solo lo considera si esta dentro de la esfera circunscrita
       * (de radio sqrt(3) / 2) */
      if (dist2 < 0.75) {
        int ny = pixely + a, nx = pixelx + b;              
        if (ny >= 0 && ny < Ydim && nx >= 0 && nx < Xdim) {
          inforayo ir;
          ir.pos = pos2d(ny, nx);
          ir.peso = kaiser_proj(sqrtf(dist2), 2.0, 3.6f, 2) - 0.97;  // 5 regs
          array[tam++] = ir;  // esto aade 4 registros

          float2 sl = slice[ir.pos];  // +2 reg.
          sl.x += ir.peso;
          sl.y += volpos * ir.peso;
          slice[ir.pos] = sl;         // 50% del tiempo de proyeccion
        }
      }
    }
  rayosporvoxel[pos].tam = tam;
}

__global__ void sumaslices(float *proyeccion, infoproyslice *slice, int numslices, int Ydim, int Xdim, int hb
#ifdef DEBUG    
    , float *debug
#endif    
    ) {
  int pos = blockIdx.x * hb + threadIdx.x;
  if (pos >= Ydim * Xdim) return;

  slice += pos;
  float proy = 0, normFactor = 0;
  for (int i = 0; i < numslices; ++i) {
    normFactor += (*slice).x;
    proy += (*slice).y;
    slice += Ydim * Xdim;
  }
  proy = (proyeccion[pos] - proy) / normFactor;
  proyeccion[pos] = proy;
}

int voxeldrivenproj(const float *proycpu, matrix3d eulerMatrix, float lambda, int hbproyecta, int hbbackproyecta) {
  // Copia la proyeccion medida (la del ImageXmipp)
  cudaMemcpy(proyeccion, proycpu, sizeof(*proyeccion) * Ydim * Xdim, cudaMemcpyHostToDevice);

  // Calcula que direccion es mejor
  int midX = (Xdim + 1) / 2, midY = (Ydim + 1) / 2, midZ = (Zdim + 1) / 2, hilos, numslices;
  double dx = fabs(eulerMatrix.val[0][2]), dy = fabs(eulerMatrix.val[1][2]), dz = fabs(eulerMatrix.val[2][2]);
  int dir;  // 0 (x), 1 (y), 2 (z)

  if (dx >= dy && dx >= dz)      { dir = 0; hilos = midZ * midY * Xdim; numslices = Xdim; }
  else if (dy >= dx && dy >= dz) { dir = 1; hilos = midZ * Ydim * midX; numslices = Ydim; }
  else                           { dir = 2; hilos = Zdim * midY * midX; numslices = Zdim; }
  
  // Transpone la matriz (para tener la que va de 3D a 2D)
  for (int i = 0; i < 3; ++i) for (int j = 0; j < i; ++j)
    swap(eulerMatrix.val[i][j], eulerMatrix.val[j][i]);
 
  cudaMemset(slice, 0, sizeof(*slice) * Zdim * Ydim * Xdim);

  // Proyecta (calcula los rayos que pasan por cada voxel)
  {
    dim3 dimGrid((hilos + hbproyecta - 1) / hbproyecta, 1);
    dim3 dimBlock(hbproyecta, 1);
    for (int k = 0; k < 4; ++k) {
      proyectavoxels<<<dimGrid, dimBlock>>>(volumen, rayosporvoxel, slice, Zdim, Ydim, Xdim,
        eulerMatrix.val[0][0], eulerMatrix.val[0][1], eulerMatrix.val[0][2],
        eulerMatrix.val[1][0], eulerMatrix.val[1][1], eulerMatrix.val[1][2],
        dir, k, hbproyecta
#ifdef DEBUG
        , debug
#endif        
        );
      if (cudaGetLastError() != cudaSuccess) return ERROR_PROYVOXEL;
    }
  }

  /* Suma los valores de los slices para calcular las proyecciones de verdad,
   * comparar con los valores medidos y normalizar
   * Se podria hacer lo de hacerlo totalmente paralelo en log n pasos,
   * pero esto es el 5% del tiempo de proyeccion asi que no vale la pena */
  {
    int hb = 64;
    dim3 dimGrid((Ydim * Xdim + hb - 1) / hb, 1);
    dim3 dimBlock(hb, 1);
    sumaslices<<<dimGrid, dimBlock>>>(proyeccion, slice, numslices, Ydim, Xdim, hb
#ifdef DEBUG        
        , debug
#endif        
        );
    if (cudaGetLastError() != cudaSuccess) return ERROR_SUMASLICE;
  }
  return 0;
}
